program Panels(input,output);
{======================================================}
{ Source Panels Numerical Method                       }
{ as described in Section 3.17 of                      }
{ Fundamentals of Aerodynamics                         }
{     by John D. Anderson, JR.                         }
{ Written by: Christopher L. Good                      }
{------------------------------------------------------}
{ Modification History                                 }
{  07Nov96  -clg-  Program started.                    }
{                  I array complete.                   }
{  08Nov96  -clg-  System of equations solved.         }
{  12Nov96  -clg-  Cp's computed.                      }
{  14Nov96  -clg-  Graphic display of CP's added.      }
{  15Nov96  -clg-  Run time selection of .cor file.    }
{  11Dec96  -clg-  Ability to remove unwanted panels   }
{                  added to code.  Program can now do  }
{                  multi-element airfoil plots.        }
{                  Maximum number of panels increased  }
{                  from 70 to 140 (array structure     }
{                  rearranged to allow more memory)    }
{  16Dec96  -clg-  Added angle of attack selection.    }
{                  Added scaling of graphical output.  }
{======================================================}

uses
  dos, winapi, graph, crt, strings;

{ ----------------------------------------------------------- }

const
  MAXSIZE    = 140;                { maximum number of panels }
{ ----------------------------------------------------------- }

type
  CoordType = record
    X,
    Y  : double;
  end;

  Panel_Record = record
    startpt,           {   Xi, Yi   }
    endpt,             { Xi+1, Yi+1 }
    midpt : Coordtype; {   xi, yi   }
    phi,               { angle to bottom of panel }
    beta  : double;      { angle to normal vector }
  end;


  ONE_DIM = array[1..MAXSIZE] of double;
  ONE_DIM_PTR = ^ONE_DIM;
  TWO_DIM = array[1..MAXSIZE] of ONE_DIM_PTR;

{ ----------------------------------------------------------- }

var
  { decomposition of system of equations variables }
  I_Array,
  Decomp1,
  Decomp2     : TWO_DIM;

  CP,
  Delta,
  Right_Side,
  Decomp1_Rt,
  Decomp2_Rt  : ONE_DIM_PTR;

  outfile,
  coordfile   : text;
  bmp_output  : file;

  Panel       : array[1..MAXSIZE] of Panel_Record;   { up to MAXSIZE panels }

  reversed,
  raw         : array[1..MAXSIZE+1] of CoordType;    { up to MAXSIZE+1 x-y pts }

  A,
  B,
  C,
  D,
  Sj,
  E,
  Kappa,
  summation,
  CP_min,
  CP_scale,
  Panel_scale,
  Plot_Scale,
  biggest_x,
  biggest_y,
  smallest_x,
  smallest_y,
  range_x,
  range_y,
  attack,
  vector_x,
  vector_y,
  magnitude   : double;

  CP_range,
  CP_zero,
  current_delta,
  current_rows,
  current_col,
  last_panel,
  current_panel,
  i,
  j,
  k,
  numpts,
  numpanels   : longint;

  cor_file,
  output_file : string[255];
  airfoil     : string[50];
  outstring   : string[5];
  inchar      : char;

  Use_Range_x,
  Use_Range_y,
  verbose,
  done        : Boolean;

{ ----------------------------------------------------------- }

function arcsin(value:double):double;
{ returns the arcsin of the input value }
begin
  arcsin := arctan(value/sqrt(1.0-sqr(value)));
end;

function sindeg(value:double):double;
{ returns the sine of the value (in degrees) }
begin
  sindeg := sin(value*PI/180.0);
end;

function cosdeg(value:double):double;
{ returns the cosine of the value (in degrees) }
begin
  cosdeg := cos(value*PI/180.0);
end;

{ ----------------------------------------------------------- }

procedure Write_BMP_Header(bmp_width    : longint;    { width must end on 32 bit boundary }
                           bmp_height   : longint;    {    = divisable by 4               }
                           var bmp_file : file);

type
  TRGBQuad = record
    rgbBlue     : Byte;
    rgbGreen    : Byte;
    rgbRed      : Byte;
    rgbReserved : Byte;
  end;
  TBitmapInfoHeader = record
    biSize          : Longint;
    biWidth         : Longint;
    biHeight        : Longint;
    biPlanes        : Word;
    biBitCount      : Word;
    biCompression   : Longint;
    biSizeImage     : Longint;
    biXPelsPerMeter : Longint;
    biYPelsPerMeter : Longint;
    biClrUsed       : Longint;
    biClrImportant  : Longint;
  end;
  TBitmapFileHeader = record
    bfType      : Word;
    bfSize      : Longint;
    bfReserved1 : Word;
    bfReserved2 : Word;
    bfOffBits   : Longint;
  end;

  BitmapFileRecord = record
    FileHeader    : TBitmapFileHeader;
    InfoHeader    : TBitmapInfoHeader;
    Colors        : array[0..15] of TRGBQuad;
  end;

var
  bitmap : BitmapFileRecord;

begin
  (*
  bitmap file setup and bitmap colors
  color table for bitmap file (printed to paper)
    0    1 ....................... 12    13     14     15
   white   greyscale (light<-dark)      black  green  red
  *)
  with bitmap.Colors[0] do           { white }
  begin
    rgbRed      := $ff;
    rgbGreen    := $ff;
    rgbBlue     := $ff;
    rgbReserved := 0;
  end;
  for i:=1 to 13 do
  with bitmap.Colors[i] do           { black --> (greyscale) }
  begin
    rgbRed      := (13-i)*18;
    rgbGreen    := (13-i)*18;
    rgbBlue     := (13-i)*18;
    rgbReserved := 0;
  end;
  with bitmap.Colors[14] do           { green }
  begin
    rgbRed      := $00;
    rgbGreen    := $ff;
    rgbBlue     := $00;
    rgbReserved := 0;
  end;
  with bitmap.Colors[15] do           { red   }
  begin
    rgbRed      := $ff;
    rgbGreen    := $00;
    rgbBlue     := $00;
    rgbReserved := 0;
  end;
  with bitmap.FileHeader do
  begin
    bfType      := 19778;
    bfSize      := trunc(bmp_width/2*bmp_height)+118;
    bfReserved1 := 0;
    bfReserved2 := 0;
    bfOffBits   := 118;
  end;
  with bitmap.InfoHeader do
  begin
    biSize          := 40;
    biWidth         := bmp_width;
    biHeight        := bmp_height;
    biPlanes        := 1;
    biBitCount      := 4;
    biCompression   := 0;
    biSizeImage     := 0;
    biXPelsPerMeter := 0;
    biYPelsPerMeter := 0;
    biClrUsed       := 16;
    biClrImportant  := 16;
  end;

  blockwrite(bmp_file,bitmap,sizeof(bitmap));
end;

{ ----------------------------------------------------------- }

procedure Write_Pixels(bmp_width    : longint;      { width must end on 32 bit boundary }
                       bmp_height   : longint;      {    = divisable by 4               }
                       x_base       : longint;
                       y_base       : longint;
                       var bmp_file : file);
type
  BitMapArray = array[0..349] of Byte;   { widest to save = (2 * 350) = 700 pixels wide }
var
  x,y         : longint;
  pixels      : BitMapArray;
begin
  for y:=0 to bmp_height-1 do
  begin
    for x:=x_base to trunc(bmp_width/2)-1 do
      pixels[x]:=(GetPixel(x*2,y_base-y) * 16) + GetPixel(x*2+1,y_base-y);
    blockwrite(bmp_file,pixels,trunc(bmp_width/2));
    Putpixel(0,y_base-y,14);
  end;
end;

{ ----------------------------------------------------------- }

procedure Initialize;
const
  graphdriver  : integer = vga;
  graphmode    : integer = vgahi;
  PathToDriver : string  = '';

var
  i           : integer;
begin
  (*
  video driver setup and screen colors
  color table for screen :
     0    1 ....................... 12    13     14     15
    black   greyscale (dark->light)      white  green  red
  *)
  DirectVideo := False;
  InitGraph(GraphDriver, GraphMode, PathToDriver);
  for i:=0 to 15 do SetPalette(i,i);
  for i:=0 to 12 do SetRGBPalette( i,i*5,i*5,i*5);   { black --> (greyscale) }
  SetRGBPalette(13,$ff,$ff,$ff);                     { white }
  SetRGBPalette(14,$00,$ff,$00);                     { green }
  SetRGBPalette(15,$ff,$00,$00);                     { red   }

end;

{ ----------------------------------------------------------- }

procedure ComputeABCDSjE(XJ  : double;    { panel[j].startpt.x }
                         XJ1 : double;    { panel[j].endpt.x }
                         YJ  : double;    { panel[j].startpt.y }
                         YJ1 : double;    { panel[j].endpt.y }
                         phiI: double;    { panel[i].phi }
                         phiJ: double;    { panel[j].phi }
                         xI  : double;    { panel[i].midpt.x }
                         yI  : double;    { panel[i].midpt.y }
                         var A : double;
                         var B : double;
                         var C : double;
                         var D : double;
                         var Sj: double;
                         var E : double);
begin
      A:= -(xI-XJ)*cos(phiJ)-(yI-YJ)*sin(phiJ);

      B:= sqr(xI-XJ)+sqr(yI-YJ);

      C:= sin(phiI-phiJ);

      D:= (yI-YJ)*cos(phiI)-(xI-XJ)*sin(phiI);

      Sj:= sqrt( sqr(XJ1-XJ)+ sqr(YJ1-YJ));

      E:= (xI-XJ)*sin(phiJ)-(yI-YJ)*cos(phiJ);
end;

{ ----------------------------------------------------------- }

begin { main }

  clrscr;
  writeln('Welcome to PANELS ! - written by Captain Chris Good, USAF');
  writeln('---------------------------------------------------------');
  writeln;
  writeln('Enter the location (path/filename.cor) of the coordinate file :');
  readln(cor_file);
  writeln;

  {$I-}
  assign(coordfile,cor_file);
  reset(coordfile);
  {$I+}
  if (IOResult<>0) or (cor_file='') then
  begin
    writeln('The input coordinate file ''',cor_file,''' does not exist.');
    exit;
  end;

  writeln('Enter the location (path/filename) where output should be placed :');
  writeln('The output files will have this filename with .TXT and .BMP as extensions.');
  readln(output_file);
  writeln;
  writeln('Text file output   -> ',output_file,'.TXT');
  writeln('Bitmap file output -> ',output_file,'.BMP');
  writeln;
  writeln('Enter the airfoil name as it will appear in the output files :');
  readln(airfoil);
  writeln;
  writeln('Enter the angle of attack ( -90.0 to 90.0 deg ) :');
  readln(attack);

  { determine if the airfoil is 'circle','circle2', or a real airfoil }
  for i:=1 to Length(airfoil) do
    airfoil[i]:=UpCase(airfoil[i]);

  biggest_x := -1000.0;
  biggest_y := -1000.0;
  smallest_x := 1000.0;
  smallest_y := 1000.0;
  { read the raw coords in 1..n order }
  i:=1;
  if eof(coordfile) then exit;
  while not eof(coordfile) do
  begin
    readln(coordfile,raw[i].x,raw[i].y);
    if ( raw[i].x > biggest_x ) then  biggest_x  := raw[i].x;
    if ( raw[i].y > biggest_y ) then  biggest_y  := raw[i].y;
    if ( raw[i].x < smallest_x ) then smallest_x := raw[i].x;
    if ( raw[i].y < smallest_y ) then smallest_y := raw[i].y;
    i:=i+1;
  end;
  numpts:=i-1;
  close(coordfile);

  { scale all coordinates x = (0...x max)  y = (0...y max) }
  for i:= 1 to numpts do
  begin
    raw[i].x := raw[i].x - smallest_x;
    raw[i].y := raw[i].y - smallest_y;
  end;

  biggest_x := -1000.0;
  biggest_y := -1000.0;
  smallest_x := 1000.0;
  smallest_y := 1000.0;
  { get new biggest and smallest values and resulting ranges }
  for i:=1 to numpts do
  begin
    if ( raw[i].x > biggest_x ) then  biggest_x  := raw[i].x;
    if ( raw[i].y > biggest_y ) then  biggest_y  := raw[i].y;
    if ( raw[i].x < smallest_x ) then smallest_x := raw[i].x;
    if ( raw[i].y < smallest_y ) then smallest_y := raw[i].y;
  end;
  range_x := biggest_x - smallest_x;
  range_y := biggest_y - smallest_y;

  Use_Range_x := False;
  Use_Range_y := False;
  if (range_y >= (0.5 * range_x)) then
  begin
    Use_Range_y := True;
    Plot_Scale := 250.0;
  end
  else
  begin
    Use_Range_x := True;
    Plot_Scale := 500.0;
  end;

  { scale all coordinates so that the largest value in either direction is 1.0 }
  for i:= 1 to numpts do
  begin
    if Use_Range_x then
    begin
      raw[i].x := raw[i].x * (1.0/range_x);
      raw[i].y := raw[i].y * (1.0/range_x);
    end;
    if Use_Range_y then
    begin
      raw[i].x := raw[i].x * (1.0/range_y);
      raw[i].y := raw[i].y * (1.0/range_y);
    end;
  end;

  { output the coordinates to screen and make sure they are in a clockwise orientation }
  done:=False;
  while (not done) do
  begin

    clrscr;                { clear text screen }
    initialize;            { opens graphics and sets colors (for screen and bitmap file) }
    SetColor(14);
    { output airfoil name and header info }
    SetTextStyle(TriplexFont,HorizDir,3);
    OutTextXY(15,100,'Airfoil : '+airfoil+' -> Source Panel Start/End Points');
    Str(attack:6:2,outstring);
    OutTextXY(15,120,'Angle of attack : '+outstring+' degrees');
    { Set Text Style for all other text }
    SetColor(13);
    SetTextStyle(SmallFont,HorizDir,0);
    SetUserCharSize(8,8,8,8);

    for i:=1 to numpts do
    begin
      Str(i,outstring);
      PutPixel (round(raw[i].x*Plot_Scale)+10,400-round(raw[i].y*Plot_Scale),13);
      OutTextXY(round(raw[i].x*Plot_Scale)+10,400-round(raw[i].y*Plot_Scale),outstring);
    end;

    OutTextXY(10,5, 'Look at the orientation of the panels. If they are in a clockwise');
    OutTextXY(10,15,'orientation, everything is fine.  If they are in a counterclockwise');
    OutTextXY(10,25,'orientation, the panels must have their order reversed.');
    OutTextXY(10,40,'Press  ''r'' to reverse the orientation of the panels,');
    OutTextXY(10,50,'or     ''d'' for done, because the orientation is fine,');
    OutTextXY(10,60,'or     ''q'' to quit.');

    { wait for user to press 'r','d','q' to leave graphics screen }
    repeat
      inchar:=Readkey;
      if inchar='d' then done:=True;   { orientation is correct }
      if inchar='q' then exit;         { quit program }
      if inchar='r' then               { reverse order of coordinates }
      begin
        i:=numpts;
        for j:=1 to numpts do
        begin
          reversed[i]:=raw[j];
          i:=i-1;
        end;
        for i:=1 to numpts do
          raw[i]:=reversed[i];
      end;

    until (inchar='d') or (inchar='r');

  end; { orientation is correct }

  closegraph;

  { create a text file with the numbered coordinates }
  writeln('Creating output text file ...');
  assign(outfile,output_file+'.TXT');
  rewrite(outfile);

  writeln(outfile,'Output file of program PANELS for airfoil : ',airfoil);
  { - this is not needed - the information is repeated below -
  writeln(outfile,'Raw coordinates :');
  writeln(outfile,'PT     raw x     raw y');
  for i:= 1 to numpts do
    writeln(outfile,i:2,' ',raw[i].x:9:6,' ',raw[i].y:9:6);
  }
  if (raw[1].x<>raw[numpts].x) or (raw[1].y<>raw[numpts].y) then
    writeln(outfile,'WARNING : First panel start point <> Last panel end point, panel analysis may be invalid !');
  writeln(outfile);

  { transform the raw coordinates to panel coordinates      }
  { number of panels = (numpts-1)                           }
  { panel(1) = from point(1) -> point(2)                    }
  { panel(2) = from point(2) -> point(3)                    }
  { ...                                                     }
  { panel(numpts-1) = from point(numpts-1) -> point(numpts) }
  writeln('Setting up panels ...');
  numpanels:=numpts-1;
  for i:=1 to numpanels do
  begin
    { set start point, end point, middle point }
    Panel[i].startpt :=raw[i];
    Panel[i].endpt   :=raw[i+1];
    vector_x:=Panel[i].endpt.x-Panel[i].startpt.x;
    vector_y:=Panel[i].endpt.y-Panel[i].startpt.y;
    Panel[i].midpt.x  :=(vector_x/2.0)+Panel[i].startpt.x;
    Panel[i].midpt.y  :=(vector_y/2.0)+Panel[i].startpt.y;

    { compute phi for this panel }
    magnitude:=sqrt(sqr(vector_x)+sqr(vector_y));
    if (magnitude=0) then { error - 0 vector }
    begin
      writeln(outfile);
      writeln(outfile,'Panel ',i,' start point = end point; 0 vector.');
      exit;
    end;
    { get angle to bottom of panel }
    if vector_x = 0 then { straight up or down }
    begin
      if vector_y > 0 then { straight up }
        Panel[i].phi := PI/2.0;
      if vector_y < 0 then { straight down }
        Panel[i].phi := PI+PI/2.0;
    end;
    if vector_x > 0 then { quadrant 1 or 4 }
    begin
      if vector_y = 0 then { straight right }
        Panel[i].phi:= 0.0;
      if vector_y > 0 then { quadrant 1 }
        Panel[i].phi:= arcsin( vector_y/magnitude );
      if vector_y < 0 then { quadrant 4 }
        Panel[i].phi:= (2.0*PI) - arcsin( -vector_y/magnitude );
    end;
    if vector_x < 0 then { quadrant 2 or 3 }
    begin
      if vector_y = 0 then { straight left }
        Panel[i].phi:= PI;
      if vector_y > 0 then { quadrant 2 }
        Panel[i].phi:= PI - arcsin( vector_y/magnitude );
      if vector_y < 0 then { quadrant 3 }
        Panel[i].phi:= PI + arcsin( -vector_y/magnitude );
    end;

    { add angle of attack to angles }
    Panel[i].phi  := Panel[i].phi - (attack * (PI/180.0));
    Panel[i].beta := Panel[i].phi + (PI/2.0);

    { make sure all angles are 0 - 360 deg }
    while ( Panel[i].phi  < 0.0         ) do Panel[i].phi  := Panel[i].phi  + ( 2.0 * PI );
    while ( Panel[i].phi  > ( 2.0 * PI) ) do Panel[i].phi  := Panel[i].phi  - ( 2.0 * PI );
    while ( Panel[i].beta < 0.0         ) do Panel[i].beta := Panel[i].beta + ( 2.0 * PI );
    while ( Panel[i].beta > ( 2.0 * PI) ) do Panel[i].beta := Panel[i].beta - ( 2.0 * PI );
  end;

  last_panel:=2;
  current_panel:=1;
  done:=False;
  while (not done) do
  begin
    { make ability to delete panels (connecting different airfoils in a multielement plot) }
    initialize;            { opens graphics and sets colors (for screen and bitmap file) }
    SetColor(14);
    { output airfoil name and header info }
    SetTextStyle(TriplexFont,HorizDir,3);
    OutTextXY(15,100,'Airfoil : '+airfoil+' -> Source Panels');
    Str(attack:6:2,outstring);
    OutTextXY(15,120,'Angle of attack : '+outstring+' degrees');
    { Set Text Style for all other text }
    SetColor(13);
    SetTextStyle(SmallFont,HorizDir,0);
    SetUserCharSize(8,8,8,8);

    for i:=1 to numpanels do
    begin
      Str(i,outstring);
      PutPixel (round(Panel[i].midpt.x*Plot_Scale)+10,400-round(Panel[i].midpt.y*Plot_Scale),13);
      OutTextXY(round(Panel[i].midpt.x*Plot_Scale)+10,400-round(Panel[i].midpt.y*Plot_Scale),outstring);
    end;

    OutTextXY(10,5, 'Look at the panels. If any are used to join separate airfoils');
    OutTextXY(10,15,'(multi-element plots), the joining panels must be removed.');
    OutTextXY(10,25,'Press the ''-'' and ''+'' keys to move to a new current panel.');
    OutTextXY(10,40,'Press  ''r'' to remove the current panel,');
    OutTextXY(10,50,'or     ''d'' for done, because the panels are correct,');
    OutTextXY(10,60,'or     ''q'' to quit.');

    { wait for user to press 'r','d','q' to leave graphics screen }
    repeat
      { make last panel line black }
      SetColor(0);
      Line(round(Panel[last_panel].startpt.x*Plot_Scale)+10,
           400-round(Panel[last_panel].startpt.y*Plot_Scale),
           round(Panel[last_panel].endpt.x*Plot_Scale)+10,
           400-round(Panel[last_panel].endpt.y*Plot_Scale));
           { make last panel label white }
      Str(last_panel,outstring);
      SetColor(13);
      OutTextXY(round(Panel[last_panel].midpt.x*Plot_Scale)+10,
                400-round(Panel[last_panel].midpt.y*Plot_Scale),outstring);
      PutPixel(round(Panel[last_panel].midpt.x*Plot_Scale)+10,
                400-round(Panel[last_panel].midpt.y*Plot_Scale),13);
      { make current panel red }
      Str(current_panel,outstring);
      SetColor(15);
      OutTextXY(round(Panel[current_panel].midpt.x*Plot_Scale)+10,
                400-round(Panel[current_panel].midpt.y*Plot_Scale),outstring);
      Line(round(Panel[current_panel].startpt.x*Plot_Scale)+10,
           400-round(Panel[current_panel].startpt.y*Plot_Scale),
           round(Panel[current_panel].endpt.x*Plot_Scale)+10,
           400-round(Panel[current_panel].endpt.y*Plot_Scale));

      inchar:=Readkey;
      last_panel:=current_panel;
      if (inchar='-') or (inchar='_') then current_panel:=current_panel-1;
      if (inchar='=') or (inchar='+') then current_panel:=current_panel+1;
      if current_panel<1 then current_panel:=numpanels;
      if current_panel>numpanels then current_panel:=1;

      if inchar='d' then done:=True;   { panels are correct }
      if inchar='q' then exit;         { quit program }
      if inchar='r' then               { remove current panel }
      begin
        if current_panel<>numpanels then
          for i:=current_panel to numpanels do
            Panel[i]:=Panel[i+1];

        numpanels:=numpanels-1;
        current_panel:=1;
      end;

    until (inchar='d') or (inchar='r');
    closegraph;
  end;

  writeln(outfile,'Angle of attack : ',attack:6:2,' degrees');
  writeln(outfile,'Panel information:');
  writeln(outfile,'PNL    start x    start y      end x      end y      mid x      mid y   ',
                  'phi(rad)  phi(deg) beta(rad) beta(deg)');
  for i:=1 to numpanels do
    writeln(outfile,i:3,' ',Panel[i].startpt.x:10:4,' ',Panel[i].startpt.y:10:4,
                        ' ',Panel[i].endpt.x:10:4,  ' ',Panel[i].endpt.y:10:4,
                        ' ',Panel[i].midpt.x:10:4,  ' ',Panel[i].midpt.y:10:4,
                        ' ',Panel[i].phi:10:6 ,' ',(Panel[i].phi*180.0/PI):8:3,
                        ' ',Panel[i].beta:10:6,' ',(Panel[i].beta*180.0/PI):8:3);

  { create I array from panel information }
  for i:= 1 to numpanels do
  begin
    writeln('Creating I_array[',i,']');
    new(I_Array[i]);
  end;
  new(Right_Side);
  for i:= 1 to numpanels do
  begin
    for j:= 1 to numpanels do
    begin
      if i=j then
        I_Array[i]^[j] := 0.0
      else
        begin
          ComputeABCDSjE(Panel[j].startpt.x,
                       Panel[j].endpt.x,
                       Panel[j].startpt.y,
                       Panel[j].endpt.y,
                       Panel[i].phi,
                       Panel[j].phi,
                       Panel[i].midpt.x,
                       Panel[i].midpt.y,
                       A,B,C,D,Sj,E);
          I_Array[i]^[j]:=(C/2.0)*ln( ( sqr(Sj)+(2.0*A*Sj)+B )/ B) + ((D-(A*C))/E) * ( arctan((Sj+A)/E) - arctan(A/E) );
        end;
    end;
  end;

  { add PI into diagonal of I array and set up system of equations }
  writeln(outfile);
  writeln(outfile,'System of equations :');
  writeln(outfile);
  write(outfile,'+-'); for i:=1 to numpanels do write(outfile,'------------');
  writeln(outfile,'+  +------------+     +---------------------------------+');

  for i:=1 to numpanels do
  begin
    Right_Side^[i]:=-cos(Panel[i].beta);
    write(outfile,'| ');
    for j:=1 to numpanels do
    begin
      if i=j then I_Array[i]^[j]:=PI;
      write(outfile,I_Array[i]^[j]:11:7,' ');
    end;
    writeln(outfile,'|  | Delta[',i:3,'] |  =  | ',Right_Side^[i]:11:7,' x 2.0 x PI x V(inf) |');
  end;
  write(outfile,'+-'); for i:=1 to numpanels do write(outfile,'------------');
  writeln(outfile,'+  +------------+     +---------------------------------+');
  writeln(outfile);

  { if number of panels is less than 10, print out all information to output file }
  verbose:=False;
  if (numpanels < 10) then verbose:=True;

  { decompose system of equations and solve for DELTA(1)-DELTA(n) }
  for i:=1 to numpanels do
  begin
    new(Decomp1[i]);
    new(Decomp2[i]);
    writeln('Creating Decomposition arrays[',i,']');
  end;
  new(Decomp1_Rt);
  new(Decomp2_Rt);
  new(Delta);
  new(CP);
  for current_delta := 1 to numpanels do
  begin
    writeln('Decomposing system of equations to get Delta(',current_delta,') ..');

    { copy I_Array into working array (decomp2) }
    for i:=1 to numpanels do
    begin
      Decomp2_Rt^[i]:=Right_Side^[i];
      for j:=1 to numpanels do
        Decomp2[i]^[j]:=I_Array[i]^[j];
    end;

    current_rows:=numpanels;
    { go through the matrix one column at a time and eliminate each coefficient }
    for current_col:= 1 to numpanels do
      if current_col<>current_delta then
      { eliminate DELTA(current_col) coefficient, when current_col<>current_Delta }
      begin

        { divide each row by its n_th coefficient }
        for i:=(numpanels-current_rows+1) to numpanels do
        begin
          Decomp1_Rt^[i]:=Decomp2_Rt^[i]/Decomp2[i]^[current_col];
          for j:=1 to numpanels do
            Decomp1[i]^[j]:=Decomp2[i]^[j]/Decomp2[i]^[current_col];
        end;

        { print out each row as a coefficient is set to 1.0 }
        if (verbose) then
        begin
          writeln(outfile);
          writeln(outfile,'Divided each row by its #',current_col,' coefficient :');
          for i:=(numpanels-current_rows+1) to numpanels do
          begin
            for j:=1 to numpanels do
            begin
              write(outfile,Decomp1[i]^[j]:11:7,'D',j,' ');
              if j<numpanels then write(outfile,'+');
            end;
            writeln(outfile,'=',Decomp1_Rt^[i]:11:7,' x 2.0 x PI x V(inf)');
          end;
        end;

        { subtract each row from the current top row to get n-1 rows }
        for i:=(numpanels-current_rows+2) to numpanels do
        begin
          Decomp2_Rt^[i]:=Decomp1_Rt^[i]-Decomp1_Rt^[numpanels-current_rows+1];
          for j:=1 to numpanels do
            Decomp2[i]^[j]:=Decomp1[i]^[j]-Decomp1[numpanels-current_rows+1]^[j];
        end;

        { print out each row as a coefficient is eliminated }
        if (verbose) then
        begin
          writeln(outfile);
          writeln(outfile,'Subtracted : new row = row n - row ',numpanels-current_rows+1,
                          ' ; DELTA(',current_col,') now eliminated :');
          for i:=(numpanels-current_rows+2) to numpanels do
          begin
            for j:=1 to numpanels do
            begin
              write(outfile,Decomp2[i]^[j]:11:7,'D',j,' ');
              if j<numpanels then write(outfile,'+');
            end;
            writeln(outfile,'=',Decomp2_Rt^[i]:11:7,' x 2.0 x PI x V(inf)');
          end;
        end;

        current_rows:=current_rows-1;
      end;
    Delta^[current_delta]:=Decomp2_Rt^[numpanels]/Decomp2[numpanels]^[current_delta];

    { This next section of code is algebraically correct, but it falls apart due
      to overflow and underflow in all computers representation of numbers.
      Even going to higher accuracy (64 bit reals), it still falls apart after
      about 20 - 25 solutions. To achieve correct values for the solutions,
      each Delta must be solved independently of the preceeding Deltas.  By
      taking advantage of the fact that a Delta has already been solved, the code
      is speed up dramatically.  Alas, it also gives invalid results. }
    {-- gives invalid results ---
    for i:=1 to numpanels do
    begin
      Right_Side^[i] := Right_Side^[i] - (I_Array[i]^[current_delta]*Delta^[current_delta]);
      I_Array[i]^[current_delta]:= 0.0;
    end;
     ---------------------------- }

    { output the final value of Delta(current_delta) }
    if (verbose) then
    begin
      writeln(outfile);
      writeln(outfile,'Delta(',current_delta,') = ',Delta^[current_delta]:11:7,' x 2.0 x PI x V(inf)');
      writeln(outfile,'-----------------------------------------------------------------------');
    end;
  end;

  { print out all deltas : delta[1]..delta[n] }
  writeln(outfile);
  for i:=1 to numpanels do
    writeln(outfile,'Delta(',i,') = ',Delta^[i]:11:7,' x 2.0 x PI x V(inf)');
  writeln(outfile,'-----------------------------------------------------------------------');

  { all done decomposing system of equations, now get pressure coefficients }
  writeln('Computing pressure coefficients ...');
  for i:=1 to numpanels do
  begin
    if (verbose) then
    begin
      writeln(outfile);
      writeln(outfile,'Computing Pressure Coefficient for Panel ',i,' :');
      writeln(outfile);
      writeln(outfile,'Kappa = ((D-AC)/(2E))*ln((Sj^2+2ASj+B)/B)-C(arctan((Sj+A)/E)-arctan(A/E))');
    end;

    summation:=0.0;
    for j:=1 to numpanels do
    begin
      ComputeABCDSjE(Panel[j].startpt.x,
                     Panel[j].endpt.x,
                     Panel[j].startpt.y,
                     Panel[j].endpt.y,
                     Panel[i].phi,
                     Panel[j].phi,
                     Panel[i].midpt.x,
                     Panel[i].midpt.y,
                     A,B,C,D,Sj,E);

      if (E=0.0) or (B=0.0) then
        Kappa:=0.0
      else
        Kappa:= ((D-A*C)/(2.0*E)) * ln( ( sqr(Sj)+(2.0*A*Sj)+B )/ B) - ( C * ( arctan((Sj+A)/E) - arctan(A/E) ) );

      summation:=summation+(Delta^[j]*Kappa);

      if (verbose) then writeln(outfile,'i=',i,' j=',j,' : A,B,C,D,Sj,E --> Kappa : ',
                    A:8:4,' ',B:8:4,' ',C:8:4,' ',D:8:4,' ',Sj:8:4,' ',E:8:4,' --> ',Kappa:8:4);
    end;


    CP^[i]:=1-sqr( sin(Panel[i].beta) + summation );

    if (verbose) then
    begin
      writeln(outfile,'Summation of : Delta(j=1..',numpanels,') x KAPPA(i=',i,',j=1..',numpanels,') = ',
                     summation:11:7,' x 2.0 x PI x V(inf)');
      writeln(outfile);
      writeln(outfile,'Beta(',i,')       = ',Panel[i].beta:10:6,' (rads) = ',(Panel[i].beta*180.0/PI):8:3,' (deg)');
      writeln(outfile,'Sin(',Panel[i].beta:8:6,') = ',sin(Panel[i].beta):11:7);
      writeln(outfile,'V(',i,')          = V(inf) x Sin(Beta(',i,')) + ( Summation of : Delta(j=1..',numpanels,
                    ') x KAPPA(i=',i,',j=1..',numpanels,') ) / 2.0 x PI');
      writeln(outfile,'--------      = -----  x ------------ + ------ x -----------');
      writeln(outfile,'V(',i,')          = V(inf) x     ',sin(Panel[i].beta):8:3,' + V(inf) x ',summation:11:7);
      writeln(outfile);
      writeln(outfile,'V(',i,') / V(inf) = ',sin(Panel[i].beta):11:7,' + ',summation:11:7);
      writeln(outfile);
      writeln(outfile,'Cp(',i,')         = 1 - sqr( V(',i,') / V(inf) )');
      writeln(outfile,'Cp(',i,')         = ',CP^[i]:10:6);
      writeln(outfile);
      writeln(outfile,'---------------------------------------------------------------------------------------');
    end;

  end;

  CP_min:=1.0;
  { print out all pressure coefficients and get minimum CP }
  writeln(outfile);
  writeln(outfile,'Panel, Panel_X, Panel_Y, CP');
  for i:=1 to numpanels do
  begin
    if CP^[i]<CP_min then CP_min:=Cp^[i];
    writeln(outfile,i:3,' , ',Panel[i].midpt.x:10:6,
                        ' , ',Panel[i].midpt.y:10:6,
                        ' , ',CP^[i]:14:8);
  end;
  close(outfile);

  { plot all pressure coefficients to graphics screen }
  initialize;            { opens graphics and sets colors (for screen and bitmap file) }
  SetColor(14);
  { output airfoil name and header info }
  SetTextStyle(TriplexFont,HorizDir,3);
  OutTextXY(15,2,'Airfoil : '+airfoil+' -> Source Panels + Cp''s');
  { Set Text Style for all other text }
  SetColor(13);
  SetTextStyle(SmallFont,HorizDir,0);
  SetUserCharSize(8,8,8,8);

  for i:=1 to numpanels do
    begin
      Str(i,outstring);
      PutPixel (round(Panel[i].midpt.x*Plot_Scale)+10,140-round(Panel[i].midpt.y*Plot_Scale),13);
      OutTextXY(round(Panel[i].midpt.x*Plot_Scale)+10,140-round(Panel[i].midpt.y*Plot_Scale),outstring);
    end;

  { plot the pressure coeficients below the airfoil }
  { plot the y axis }
  Line(25,160,25,460);      { left y axis }
  Line(555,160,555,460);    { right y axis }
  { compute scaling factor due to range of Cp's }
  CP_range:=1-(trunc(CP_min)-1);
  CP_scale:=300.0/CP_range;
  CP_zero:=160+round(CP_scale);
  { plot the x axis }
  Line(25,CP_zero,555,CP_zero);

  { print the values of CP : 1..CP_min -> label the y axis }
  for i:=1 downto (trunc(CP_min)-1) do
  begin
    Str(i,outstring);
    OutTextXY(2,CP_zero-round(i*CP_scale)-5,outstring);
    OutTextXY(560,CP_zero-round(i*CP_scale)-5,outstring);
    Line(23,CP_zero-round(i*CP_scale),27,CP_zero-round(i*CP_scale));
    Line(553,CP_zero-round(i*CP_scale),557,CP_zero-round(i*CP_scale));
  end;

  { plot the values of CP }
  Panel_scale:=520.0/(numpanels-1);
  for i:=1 to numpanels do
  begin
    SetColor(13);
    Line(30+round((i-1)*Panel_scale),CP_zero-3,30+round((i-1)*Panel_scale),CP_zero+3);
    SetColor(15);
    Circle(30+round((i-1)*Panel_scale),CP_zero-round(CP^[i]*CP_scale),3);
  end;

  SetColor(13);
  { wait for user to press 's','q' to leave graphics screen }
  OutTextXY(10,466,'Press ''s'' to save screen, ''q'' to quit.');
  repeat
    inchar:=Readkey;
    if inchar='q' then exit;  {bail out}
  until (inchar='s');

  { save screen as bitmap file }
  OutTextXY(250,466,'Saving this screen to a bitmap (.bmp) file.');
  assign(bmp_output,output_file+'.BMP');
  rewrite(bmp_output,1);
  Write_BMP_Header(576,465,bmp_output);
  Write_Pixels(576,465,0,465,bmp_output);
  { this will save the screen from x=0..559, y=465..1 }
  close(bmp_output);
  closegraph;

end.
